#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : DNN
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-09-14 11:52
   Description : description what the main function of this file
"""
from stensorflow.ml.nn.layers.layer import Layer
from stensorflow.ml.nn.networks.NN import NN
from typing import List
from stensorflow.global_var import StfConfig
from stensorflow.basic.basic_class.private import PrivateTensor
from stensorflow.basic.basic_class.pair import SharedPair
from stensorflow.ml.nn.layers.input import Input
from stensorflow.ml.nn.layers.dense import Dense
from stensorflow.ml.nn.layers.loss import Loss, BinaryCrossEntropyLossWithSigmoid, CrossEntropyLossWithSoftmax
from stensorflow.ml.nn.layers.relu import ReLU
from stensorflow.random.random import random_init


class DNN(NN):
    def __init__(self, feature: PrivateTensor, label: PrivateTensor, dense_dims: List[int],
                 feature_another: PrivateTensor = None, loss=None):
        """
        Full-connected Neural Network. Must have feature.owner != label.owner. If
        feature_another is not None, feature_another.owner==label.owner.
        :param feature:            feature in different party of label.owner
        :param label:
        :param dense_dims:  [d0, d1, ..., dn]  where d0=dim(feature) dn=dim(label), di is the
        dimension of i-th hidden layer
        :param feature_another:    feature in same party of label.owner
        :param loss
        """

        if len(dense_dims) < 2:
            raise Exception("must have len(dense_dims)>=2")
        if feature_another is None:
            self.double_feature = False
            if feature.shape[1] != dense_dims[0]:
                raise Exception("must have x.shape[1] == dense_dims[0]")
            super(DNN, self).__init__()
            layer = Input(dim=dense_dims[0], x=feature)
            self.addLayer(ly=layer)
            input_layers = [layer]

        else:
            self.double_feature = True
            if feature.shape[1] + feature_another.shape[1] != dense_dims[0]:
                raise Exception("must have x.shape[1] == dense_dims[0]")
            if feature.owner == feature_another.owner:
                raise Exception("must have feature.owner != feature_another.owner")
            super(DNN, self).__init__()
            input_layer = Input(dim=feature.shape[1], x=feature)
            self.addLayer(ly=input_layer)
            input_layer_another = Input(dim=feature_another.shape[1], x=feature_another)
            self.addLayer(input_layer_another)
            input_layers = [input_layer, input_layer_another]

        for i in range(1, len(dense_dims)):
            if i == 1:
                layer = Dense(output_dim=dense_dims[i], fathers=input_layers)
            else:
                layer = Dense(output_dim=dense_dims[i], fathers=[layer])
            self.addLayer(ly=layer)
            if i < len(dense_dims) - 1:
                layer = ReLU(output_dim=dense_dims[i], fathers=[layer])
                self.addLayer(ly=layer)

        layer_label = Input(dim=dense_dims[-1], x=label)
        self.addLayer(ly=layer_label)

        if loss is None or loss == "BinaryCrossEntropyLossWithSigmoid" or \
                loss == Loss.BinaryCrossEntropyLossWithSigmoid:
            layer_loss = BinaryCrossEntropyLossWithSigmoid(layer_score=layer, layer_label=layer_label)
            self.addLayer(ly=layer_loss)
        elif loss == "CrossEntropyLossWithSoftmax" or loss == Loss.CrossEntropyLossWithSoftmax:
            layer_loss = CrossEntropyLossWithSoftmax(layer_score=layer, layer_label=layer_label)
            self.addLayer(ly=layer_loss)
        else:
            raise Exception("unsupposed loss")

    def predict(self, x, x_another=None, out_prob=True) -> SharedPair:
        self.cut_off()
        l_input = self.layers[0]
        assert isinstance(l_input, Input)
        l_input.replace(x)
        self.layers[0] = l_input

        if self.double_feature:
            l_input_another = self.layers[1]
            if not isinstance(l_input_another, Input):
                raise Exception("l_input_another mast be a Input layer")
            l_input_another.replace(x_another)
            self.layers[1] = l_input_another

        ly = self.layers[-1]
        if not isinstance(ly, Layer):
            raise Exception("l must be a Layer")
        else:
            ly.forward()

        if out_prob:
            return ly.y
        else:
            return ly.score

    def predict_to_file(self, sess, x, predict_file_name,
                        batch_num, idx, model_file_machine="R", record_num_ceil_mod_batch_size=0,
                        x_another=None, with_sigmoid=True):
        y_pred = self.predict(x=x, x_another=x_another, out_prob=with_sigmoid)

        id_y_pred = y_pred.to_tf_str(owner=model_file_machine, id_col=idx)
        random_init(sess)

        with open(predict_file_name, "w") as f:
            for batch in range(batch_num - 1):
                records = sess.run(id_y_pred)
                records = "\n".join(records.astype('str'))
                f.write(records + "\n")

            records = sess.run(id_y_pred)[0:record_num_ceil_mod_batch_size]
            records = "\n".join(records.astype('str'))
            f.write(records + "\n")

    def print(self, sess, model_file_machine="R"):
        for ly in self.layers:
            if isinstance(ly, Dense):
                for weight in ly.w:
                    weight = weight.to_tf_tensor(model_file_machine)
                    print(sess.run(weight))



from stensorflow.ml.nn.networks.NN import NN
from stensorflow.ml.nn.layers.layer import Layer
from stensorflow.ml.nn.layers.input import Input
from stensorflow.ml.nn.layers.relu import ReLU
from stensorflow.ml.nn.layers.conv2d import Conv2d
from stensorflow.ml.nn.layers.pooling import AveragePooling2D
from stensorflow.ml.nn.layers.loss import CrossEntropyLossWithSoftmax
from stensorflow.ml.nn.layers.flatten import Flatten
from stensorflow.ml.nn.layers.dense import Dense, Dense_Local
from stensorflow.basic.basic_class.pair import SharedPair
from stensorflow.random import random
import numpy as np


class SIMPLE_CNN(NN):
    def __init__(self, feature, label, loss=None):
        super(SIMPLE_CNN, self).__init__()
        # input layer, init data；
        # 这里将dim设置位输入的wight,后续不使用；仅仅是为了应用原有的模板
        layer = Input(dim=28, x=feature)
        local_layer_owner = layer.owner
        self.addLayer(layer)
        # convolutional layer with 1 input channel, 16 output channels and a 5×5 filter
        layer = Conv2d(output_dim=None, fathers=[layer], filters=16,
                       kernel_size=5, input_shape=[28, 28, 1])
        # layer = Conv2d_Local(output_dim=28, fathers=[layer], filters=16,
        #                kernel_size=5, input_shape=[28, 28, 1],owner=local_layer_owner)
        self.addLayer(layer)

        # Relu Layer
        layer = ReLU(output_dim=layer.output_dim, fathers=[layer])
        # layer = ReLU_Local(output_dim=28, fathers=[layer], owner=local_layer_owner)
        self.addLayer(layer)

        # Average pool
        layer = AveragePooling2D(output_dim=None, fathers=[layer], pool_size=(2, 2))
        # layer = AveragePooling2D_Local(output_dim=None, fathers=[layer], pool_size=(2, 2), owner=local_layer_owner)
        self.addLayer(layer)

        # flatten data, only consider data_format = "NWHC"
        # 这里需要给出正确的output_dim，方便后续的全连接层
        layer = Flatten(output_dim=None, fathers=[layer])
        # layer = Flatten_Local(output_dim=2304, fathers=[layer], owner=local_layer_owner)
        self.addLayer(layer)
        # 临时添加测试
        # layer = Dense(output_dim=100, fathers=[layer])
        # self.addLayer(layer)
        # layer = ReLU(output_dim=100, fathers=[layer])
        # self.addLayer(layer)
        # a 256 × 10 linear layer
        # layer = Dense(output_dim=10, fathers=[layer])
        layer = Dense_Local(output_dim=10, fathers=[layer], owner=local_layer_owner)
        self.addLayer(layer)
        # 输出层
        layer_label = Input(dim=10, x=label)
        self.addLayer(ly=layer_label)
        # 损失计算
        layer_loss = CrossEntropyLossWithSoftmax(layer_score=layer, layer_label=layer_label)
        self.addLayer(ly=layer_loss)

    def predict(self, x, out_prob=True):
        self.cut_off()
        # 输入层
        l_input = self.layers[0]
        assert isinstance(l_input, Input)
        l_input.replace(x)
        self.layers[0] = l_input
        # 输出层
        ly = self.layers[-1]
        if not isinstance(ly, Layer):
            raise Exception("l must be a Layer")
        else:
            ly.forward()

        if out_prob:
            return ly.y
        else:
            return ly.score

    def predict_to_file(self, sess, x, predict_file_name,
                        pred_batch_num, model_file_machine,
                        record_num_ceil_mod_batch_size,
                        with_sigmoid):
        y_pred = self.predict(x=x,  out_prob=with_sigmoid)
        id_y_pred = y_pred.to_tf_str(owner=model_file_machine)
        random.random_init(sess)
        # sess.run(tf.compat.v1.global_variables_initializer())
        # with open(predict_file_name, "w") as f:
        #     for batch in range(batch_num - 1):
        #         records = sess.run(id_y_pred)
        #         records = "\n".join(records.astype('str'))
        #         f.write(records + "\n")
        #
        #     records = sess.run(id_y_pred)[0:record_num_ceil_mod_batch_size]
        #     records = "\n".join(records.astype('str'))
        #     f.write(records + "\n")
        # 分批写入文件
        with open(predict_file_name, "w") as f:
            for batch in range(pred_batch_num):
                records = sess.run(id_y_pred)
                records = "\n".join(records.astype('str'))
                # records.to_file()
                f.write(records + "\n")

    def print(self, model_file_machine, sess):
        for ly in self.layers:
            if isinstance(ly, Dense) or isinstance(ly, Conv2d):
                print(ly)
                for weight in ly.w:
                    print(weight)
                    weight = weight.to_tf_tensor(owner=model_file_machine)
                    print("***")
                    print(sess.run(weight))

    def replace_weight(self, keras_weight):
        i = 0
        for ly in self.layers:
            if isinstance(ly, Conv2d):
                kernel = SharedPair(ownerL="L", ownerR="R", shape=keras_weight[i].shape)
                kernel.load_from_numpy(keras_weight[i])
                ly.w[0] = kernel
                # 分割
                # kernel = ly.w[0]
                # assert isinstance(kernel, SharedVariablePair)
                # kernel.load_from_tf_tensor(keras_weight[i])
                i += 1
            if isinstance(ly, Dense):
                kernel1 = SharedPair(ownerL="L", ownerR="R", shape=keras_weight[i].shape)
                kernel1.load_from_numpy(keras_weight[i])
                ly.w[0] = kernel1
                kernel2 = SharedPair(ownerL="L", ownerR="R", shape=keras_weight[i+1].shape)
                kernel2.load_from_numpy(keras_weight[i+1])
                ly.w[1] = kernel2
                # 分割
                # kernel1 = ly.w[0]
                # assert isinstance(kernel1, SharedVariablePair)
                # kernel1.load_from_tf_tensor(keras_weight[i])
                # kernel2 = ly.w[1]
                # assert isinstance(kernel2, SharedVariablePair)
                # kernel2.load_from_tf_tensor(keras_weight[i+1])
                i += 2

    def save_model(self, sess, save_file_path, model_file_machine):
        res = []
        for ly in self.layers:
            if isinstance(ly, Dense) or isinstance(ly, Conv2d):
                for weight in ly.w:
                    weight = weight.to_tf_tensor(owner=model_file_machine)
                    weight = sess.run(weight)
                    res.append(weight)
        res = np.array(res)
        np.savez(save_file_path, weight=res)


    def compare(self, keras_weight, model_file_machine, sess):
        i = 0
        for ly in self.layers:
            if isinstance(ly, Conv2d):
                # 转为tensor
                w = ly.w[0].to_tf_tensor(owner=model_file_machine)
                # 转为numpy
                w = sess.run(w)
                # 比较两个numpy的差距
                res = np.abs(w-keras_weight[i])
                res = res > 1
                if res.any():
                    print("conv2d 发生变化")
                i += 1

            if isinstance(ly, Dense):
                w1 = ly.w[0].to_tf_tensor(owner=model_file_machine)
                w1 = sess.run(w1)
                res1 = np.abs(w1-keras_weight[i])
                res1 = res1 > 5
                if res1.any():
                    print("dense 发生变化")
                i += 2
        pass


import tensorflow as tf
from stensorflow.basic.basic_class.base import SharedTensorBase
import numpy as np
from typing import Union
from stensorflow.exception.exception import StfTypeException, StfDTypeException


class SharedTensor(SharedTensorBase):
    def __repr__(self) -> str:
        return 'SharedTensor(module={}, shape={})'.format(self.module, self.inner_value.shape)


class SharedVariable(SharedTensor):
    def __init__(self, initial_inner_value=None, module: int = None, shape=None):

        if initial_inner_value is not None:
            super(SharedVariable, self).__init__(module=module, shape=initial_inner_value.shape)
            if module is None:
                self.inner_value = tf.Variable(initial_value=initial_inner_value)
            else:
                self.inner_value = tf.Variable(initial_value=initial_inner_value % module)
        elif shape is not None:
            super(SharedVariable, self).__init__(module=module, shape=shape)

    def __repr__(self) -> str:
        return 'SharedVariable(module={}, shape={})'.format(self.module, self.inner_value.shape)

    def assign(self, other):
        self.check_module_equal(other)
        assign_op = self.inner_value.assign(value=0 * self.inner_value + other.inner_value, read_value=False)
        return assign_op


def sin2pi(x: SharedTensor, fixedpoint, k: Union[int, tf.Tensor] = None):
    """
    sin 2 k pi (x/ (2**fixedpoint))
    :param x:
    :param fixedpoint:
    :param k:
    :return:
    """
    if k is None:
        k = 1
    if isinstance(k, int):
        pass
    elif isinstance(k, tf.Tensor):
        if k.dtype in [tf.dtypes.int8, tf.dtypes.int16, tf.dtypes.int32, tf.dtypes.int64]:
            k = tf.cast(k, 'int64')
        else:
            raise StfDTypeException("k", "tf.dtypes.int8, tf.dtypes.int16, tf.dtypes.int32, tf.dtypes.int64",
                                    k.dtype)
    else:
        raise StfTypeException("k", "int or tf.Tensor", type(k))
    n = (1 << fixedpoint)
    return tf.sin(
        2 * np.pi * tf.cast(k * (x.inner_value % n) % n, 'float64') / n)


def cos2pi(x: SharedTensor, fixedpoint, k: Union[int, tf.Tensor] = None):
    """
    cos 2 k pi x/ (2**fixedpoint)
    :param x:
    :param fixedpoint:
    :param k:
    :return:
    """
    if k is None:
        k = 1
    if isinstance(k, int):
        pass
    elif isinstance(k, tf.Tensor):
        if k.dtype in [tf.dtypes.int8, tf.dtypes.int16, tf.dtypes.int32, tf.dtypes.int64]:
            k = tf.cast(k, 'int64')
        else:
            raise StfDTypeException("k", "tf.dtypes.int8, tf.dtypes.int16, tf.dtypes.int32, tf.dtypes.int64",
                                    k.dtype)
    else:
        raise StfTypeException("k", "int or tf.Tensor", type(k))
    n = (1 << fixedpoint)
    return tf.cos(
        2 * np.pi * tf.cast(k * (x.inner_value % n) % n, 'float64') / n)




#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : sigmoid
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-05-14 11:42
   Description : description what the main function of this file
"""

from stensorflow.basic.basic_class.private import PrivateTensor
from stensorflow.basic.basic_class.pair import SharedPair
from stensorflow.basic.basic_class.share import sin2pi as sin2pi_share, cos2pi as cos2pi_share
import tensorflow as tf
from stensorflow.global_var import StfConfig
from typing import Union
import numpy as np


def sigmoid_poly(x: SharedPair):
    """A Chebyshev polynomial approximation of the sigmoid function."""

    w0 = 0.5
    w1 = 0.2159198015
    w3 = -0.0082176259
    w5 = 0.0001825597
    w7 = -0.0000018848
    w9 = 0.0000000072

    x1 = x
    x2 = (x1 * x).dup_with_precision(x.fixedpoint)
    x3 = (x2 * x).dup_with_precision(x.fixedpoint)
    x5 = (x2 * x3).dup_with_precision(x.fixedpoint)
    x7 = (x2 * x5).dup_with_precision(x.fixedpoint)
    x9 = (x2 * x7).dup_with_precision(x.fixedpoint)

    y1 = w1 * x1
    y3 = w3 * x3
    y5 = w5 * x5
    y7 = w7 * x7
    y9 = w9 * x9

    z = y9 + y7 + y5 + y3 + y1 + tf.constant(w0)

    return z


def sigmoid_poly_minmax(x: SharedPair):
    """A minmax polynomial approximation of the sigmoid function.
    """

    w0 = 0.5
    w1 = 0.197
    w3 = -0.004
    x1 = x
    x2 = x1 * x
    x3 = x2 * x
    y1 = w1 * x1
    y3 = w3 * x3

    z = y3 + y1 + tf.constant(w0)
    # z = y7 + y5 + y3 + y1 + w0

    return z


def sin2pi(x: SharedPair, T: int = 1, k: Union[int, tf.Tensor] = None) -> SharedPair:
    # sin(2kpix/T)
    # print("x.xL.shape=", x.xL.shape)
    # print("x.xR.shape=", x.xR.shape)
    n = int(np.log2(T))

    if 1 << n != T:
        raise Exception("T must be a power of 2")

    if k is None:
        k = 1

    if isinstance(k, int):
        pass
    elif isinstance(k, tf.Tensor):
        if k.dtype in [tf.dtypes.int8, tf.dtypes.int16, tf.dtypes.int32, tf.dtypes.int64]:
            pass
    else:
        raise Exception("the type of k is error")

    with tf.device(x.ownerL):
        yL = tf.stack([sin2pi_share(x.xL, x.fixedpoint + n, k), cos2pi_share(x.xL, x.fixedpoint + n, k)], axis=-1)

    with tf.device(x.ownerR):
        yR = tf.stack([cos2pi_share(x.xR, x.fixedpoint + n, k), sin2pi_share(x.xR, x.fixedpoint + n, k)], axis=-1)

    if StfConfig.parties == 3:
        with tf.device(x.ownerL):
            yL = tf.expand_dims(yL, axis=-2)

            zL = PrivateTensor(owner=x.ownerL)
            zL.load_from_tf_tensor(yL)
        with tf.device(x.ownerR):
            yR = tf.expand_dims(yR, axis=-1)

            zR = PrivateTensor(owner=x.ownerR)
            zR.load_from_tf_tensor(yR)

        result = (zL @ zR).dup_with_precision(new_fixedpoint=StfConfig.default_fixed_point)
        result = result.squeeze(axis=[-1, -2])
    else:
        with tf.device(x.ownerL):
            zL = PrivateTensor(owner=x.ownerL)
            zL.load_from_tf_tensor(yL)
        with tf.device(x.ownerR):
            zR = PrivateTensor(owner=x.ownerR)
            zR.load_from_tf_tensor(yR)

        result = (zL * zR).reduce_sum(axis=[-1]).dup_with_precision(new_fixedpoint=StfConfig.default_fixed_point)
    return result












def sigmoid_sin(x: SharedPair, M=16):
    """Fourier series approximation of the sigmoid function.
    https://arxiv.org/pdf/2109.11726.pdf """
    if 1 << int(np.log2(M)) != M:
        raise Exception("M must be a power of 2")
    term = 6
    sample_num = 256
    X = np.linspace(-M, M, sample_num, endpoint=False)  # -M to+M的256个值

    sigmoid = 1 / (1 + np.exp(-X))
    sm5 = sigmoid - 0.5

    sm5_odd = sm5 * 1.0
    sm5_odd[0] = 0
    F = np.fft.fft(sm5_odd)
    a = F[0:term].imag
    # a = tf.constant(a, dtype='float32', shape=[term]+[1]*len(x.shape))
    a = np.reshape(a, newshape=[term] + [1] * len(x.shape))
    a = tf.constant(a, dtype='float32')
    integers = np.reshape(range(term), newshape=[term] + [1] * len(x.shape))
    integers = tf.constant(integers, dtype='int64')
    x = x.expend_dims(axis=[0])
    s = sin2pi(x - M, T=2 * M, k=integers)
    y = -a / 128 * s
    y = y.reduce_sum(axis=[0])
    y = y + 0.5
    return y




def sigmoid_idea(x: SharedPair, M=16):
    """The idea sigmoid"""
    y=tf.sigmoid(x.to_tf_tensor("R"))
    z = SharedPair(ownerL="L", ownerR="R", shape=y.shape)
    z.load_from_tf_tensor(y)
    return z

def sigmoid_local(x: PrivateTensor):
    z = PrivateTensor(owner=x.owner)
    with tf.device(x.owner):
        y = tf.sigmoid(x.to_tf_tensor())
        z.load_from_tf_tensor(y)
    return z



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : pair.py
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-06-09 15:49
   Description : description what the main function of this file
"""
from stensorflow.basic.operator.arithmetic import add, sub, mul, matmul, rmul, rmatmul, truediv
from stensorflow.basic.operator.order import greater, geq, less, leq
from stensorflow.basic.basic_class.share import SharedVariable
from stensorflow.basic.basic_class.base import SharedPairBase, SharedTensorBase, PrivateTensorBase, get_device
import tensorflow as tf
from stensorflow.exception.exception import StfException
from stensorflow.global_var import StfConfig
import numpy as np


class SharedPair(SharedPairBase):

    def __init__(self, ownerL, ownerR, xL: SharedTensorBase = None, xR: SharedTensorBase = None,
                 fixedpoint: int = StfConfig.default_fixed_point,
                 shape=None, op_map={}):
        super(SharedPair, self).__init__(ownerL=ownerL, ownerR=ownerR, xL=xL, xR=xR, fixedpoint=fixedpoint, shape=shape)
        self.op_map = {"add": add, "sub": sub, "mul": mul, "matmul": matmul, "rmul": rmul,
                       "rmatmul": rmatmul,
                       "truediv": truediv, "gt": greater, "ge": geq, "lt": less, "le": leq}
        self.op_map.update(op_map)

    def __repr__(self) -> str:
        return 'SharedPair(ownerL={},ownerR={}, fixedpoint={}, module={}, shape={}, xL={}, xR={})'.format(
            self.ownerL.to_string(), self.ownerR.to_string(), self.fixedpoint, self.xL.module, self.xL.shape, self.xL,
            self.xR)

    def __getitem__(self, index):
        xL = self.xL[index]
        xR = self.xR[index]
        z = SharedPair(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR,
                       fixedpoint=self.fixedpoint, op_map=self.op_map)
        return z


    @classmethod
    def from_SharedPairBase(cls, x: SharedPairBase, op_map={}):
        return SharedPair(ownerL=x.ownerL, ownerR=x.ownerR, xL=x.xL, xR=x.xR, fixedpoint=x.fixedpoint, op_map=op_map)

    def _align_owner(self, other):
        "See SharedPairBase._align_owner."
        result = super(SharedPair, self)._align_owner(other)
        return self.from_SharedPairBase(result)

    def ones_like(self):
        "See SharedPairBase.ones_like."
        result = SharedPairBase.ones_like(self)
        return self.from_SharedPairBase(result, self.op_map)

    def lower_bound_like(self):
        "Return a SharedPair of value=StfConfig.lower_bd_int64"
        with tf.device(self.ownerL):
            xL = (StfConfig.lower_bd_int64 + 1) * self.xL.ones_like()
        with tf.device(self.ownerR):
            xR = - self.xR.ones_like()
        x = SharedPair(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR, fixedpoint=self.fixedpoint,
                       op_map=self.op_map)
        return x

    def upper_bound_like(self):
        with tf.device(self.ownerL):
            xL = (StfConfig.upper_bd_int64 + 1) * self.xL.ones_like()
        with tf.device(self.ownerR):
            xR = - self.xR.ones_like()
        x = SharedPair(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR, fixedpoint=self.fixedpoint,
                       op_map=self.op_map)
        return x

    def zeros_like(self):
        y = SharedPairBase.zeros_like(self)
        return SharedPair.from_SharedPairBase(y, self.op_map)

    def ones_like(self):
        y = SharedPairBase.ones_like(self)
        return SharedPair.from_SharedPairBase(y, self.op_map)

    def dup_with_precision(self, new_fixedpoint):
        result = SharedPairBase.dup_with_precision(self, new_fixedpoint)
        return self.from_SharedPairBase(result, self.op_map)

    def split(self, size_splits, axis: int = 0, num=None):
        with tf.device(self.ownerL):
            xLs = self.xL.split(size_splits=size_splits, axis=axis, num=num)
        with tf.device(self.ownerR):
            xRs = self.xR.split(size_splits=size_splits, axis=axis, num=num)
        xs = []
        for xL, xR in zip(xLs, xRs):
            xs += [SharedPair(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR, fixedpoint=self.fixedpoint,
                             op_map=self.op_map)]

        return tuple(xs)

    def concat(self, other, axis):
        if isinstance(other, SharedPairBase):
            result = super(SharedPair, self).concat(other, axis)
            return SharedPair.from_SharedPairBase(result, op_map=self.op_map)
        elif isinstance(other, PrivateTensorBase):  # can not import PrivateTensor
            if other.owner == self.ownerL:
                oppo_owner = self.ownerR
            elif other.owner == self.ownerR:
                oppo_owner = self.ownerL
            else:
                raise Exception("must have other.owner in {self.ownerL, self.ownerR}")
            other = other.to_SharedPair(other_owner=oppo_owner)
            return self.concat(other, axis)
        else:
            raise Exception("other must be SharedPairBase or PrivateTensor.")

    def stack(self, other, axis):
        if isinstance(other, SharedPairBase):
            result = super(SharedPair, self).stack(other, axis)
            return SharedPair.from_SharedPairBase(result, op_map=self.op_map)
        elif isinstance(other, PrivateTensorBase):
            if other.owner == self.ownerL:
                oppo_owner = self.ownerR
            elif other.owner == self.ownerR:
                oppo_owner = self.ownerL
            else:
                raise Exception("must have other.owner in {self.ownerL, self.ownerR}")
            other = other.to_SharedPair(other_owner=oppo_owner)
            return self.stack(other, axis)
        else:
            raise Exception("other must be SharedPairBase or PrivateTensor.")

    def transpose(self):
        with tf.device(self.ownerL):
            xL_t = self.xL.transpose()
            xR_t = self.xR.transpose()
        return SharedPair(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL_t, xR=xR_t, fixedpoint=self.fixedpoint,
                          op_map=self.op_map)

    def to_SharedPair(self, ownerL, ownerR):
        ownerL = get_device(ownerL)
        ownerR = get_device(ownerR)
        if ownerL == self.ownerL and ownerR == self.ownerR:
            return self
        elif ownerL == self.ownerR and ownerR == self.ownerL:
            return self.mirror()
        else:
            raise StfException("must have Set(ownerL, ownerR)==Set(self.ownerL, self.ownerR), "
                               "but ownerL={}, ownerR={}, self.ownerL={}, self.ownerR={}"
                               .format(ownerL, ownerR, self.ownerL, self.ownerR))

    def reduce_mean(self, axis, keepdims=False):
        with tf.device(self.ownerL):
            xL = self.xL.reduce_mean(axis=axis, keepdims=keepdims)
        with tf.device(self.ownerR):
            xR = self.xR.reduce_mean(axis=axis, keepdims=keepdims)
        return SharedPair(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR,
                          fixedpoint=self.fixedpoint, op_map=self.op_map)

    def reduce_sum(self, axis, keepdims=False):
        with tf.device(self.ownerL):
            xL = self.xL.reduce_sum(axis=axis, keepdims=keepdims)
        with tf.device(self.ownerR):
            xR = self.xR.reduce_sum(axis=axis, keepdims=keepdims)
        return SharedPair(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR,
                          fixedpoint=self.fixedpoint, op_map=self.op_map)

    def squeeze(self, axis=None):
        with tf.device(self.ownerL):
            xL = self.xL.squeeze(axis=axis)
        with tf.device(self.ownerR):
            xR = self.xR.squeeze(axis=axis)
        return SharedPair(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR, fixedpoint=self.fixedpoint,
                          op_map=self.op_map)

    def expend_dims(self, axis):
        with tf.device(self.ownerL):
            xL = self.xL.expand_dims(axis=axis)
        with tf.device(self.ownerR):
            xR = self.xR.expand_dims(axis=axis)
        return SharedPair(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR, fixedpoint=self.fixedpoint,
                          op_map=self.op_map)

    def reshape(self, shape):
        """
        See tf.reshape.
        :param shape:
        :return:
        """
        with tf.device(self.ownerL):
            xL = self.xL.reshape(shape)
        with tf.device(self.ownerR):
            xR = self.xR.reshape(shape)
        return SharedPair(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR,
                          fixedpoint=self.fixedpoint, op_map=self.op_map)


    def __add__(self, other):
        # return super(PrivateTensor, self).__add__(other)
        result = self.op_map['add'](self, other)
        return self.from_SharedPairBase(result, self.op_map)

    def __radd__(self, other):
        result = self.op_map['add'](self, other)
        return self.from_SharedPairBase(result, self.op_map)

    def __sub__(self, other):
        # return super(PrivateTensor, self).__sub__(other)
        result = self.op_map['sub'](self, other)
        return self.from_SharedPairBase(result, self.op_map)

    def __rsub__(self, other):
        result = self.op_map['sub'](other, self)
        return self.from_SharedPairBase(result, self.op_map)

    def __mul__(self, other):
        result = self.op_map['mul'](self, other)
        return self.from_SharedPairBase(result, self.op_map)

    def __rmul__(self, other):
        result = self.op_map['rmul'](other, self)
        return self.from_SharedPairBase(result, self.op_map)

    def __matmul__(self, other):
        result = self.op_map['matmul'](self, other)
        return self.from_SharedPairBase(result, self.op_map)

    def __rmatmul__(self, other):
        result = self.op_map['rmatmul'](other, self)

        return self.from_SharedPairBase(result, self.op_map)

    def __lt__(self, other):
        result = self.op_map['lt'](self, other)
        return self.from_SharedPairBase(result, self.op_map)

    def __le__(self, other):
        result = self.op_map['le'](self, other)
        return self.from_SharedPairBase(result, self.op_map)

    def __gt__(self, other):
        result = self.op_map['gt'](self, other)
        return self.from_SharedPairBase(result, self.op_map)

    def __ge__(self, other):
        result = self.op_map['ge'](self, other)
        return self.from_SharedPairBase(result, self.op_map)

    def __pow__(self, power: int):
        if not isinstance(power, int):
            raise Exception("power must be a non-negative integral number.")
        elif power < 0:
            raise Exception("power must be a non-negative integral number.")
        elif power == 0:
            return self.ones_like()
        elif power == 1:
            return self
        elif power == 2:
            return (self * self).dup_with_precision(new_fixedpoint=self.fixedpoint)
        else:  # power > 2:
            power_helf = power // 2
            y_power_helf = self ** power_helf
            y = y_power_helf * y_power_helf
            if power % 2 == 0:
                return y.dup_with_precision(new_fixedpoint=self.fixedpoint)
            else:
                return (y * self).dup_with_precision(new_fixedpoint=self.fixedpoint)

    def __truediv__(self, other):
        result = self.op_map['truediv'](self, other)
        return self.from_SharedPairBase(result, self.op_map)

    def __invert__(self):
        """
         only true for self>0
        :return:
        """
        x = 0.00005
        for _ in range(StfConfig.invert_iter_num):
            x = 2 * x - x * x * self
        return x

    def load_from_numpy(self, x: np.ndarray):
        with tf.device(self.ownerL):
            px = PrivateTensorBase(owner=self.ownerL)

            px.load_from_tf_tensor(tf.constant(x))
            xL = px.inner_value + tf.ones_like(px.inner_value)
            self.xL = SharedTensorBase(inner_value=xL, module=px.module)
            self.fixedpoint = px.fixedpoint
        with tf.device(self.ownerR):
            xR = -tf.ones_like(xL)
            self.xR = SharedTensorBase(inner_value=xR, module=px.module)


class SharedVariablePair(SharedPair):

    def __init__(self, ownerL, ownerR, xL: SharedTensorBase = None, xR: SharedTensorBase = None, fixedpoint: int = 14,
                 op_map={}, shape=None):
        super(SharedVariablePair, self).__init__(ownerL=ownerL, ownerR=ownerR, fixedpoint=fixedpoint,
                                                 op_map=op_map, shape=shape)
        if xL is not None and xR is not None:
            with tf.device(ownerL):
                self.xL = SharedVariable(initial_inner_value=xL.inner_value, module=xL.module)
            with tf.device(ownerR):
                self.xR = SharedVariable(initial_inner_value=xR.inner_value, module=xR.module)

    def __repr__(self) -> str:
        return 'SharedPairVariable(ownerL={}, ownerR={}, fixedpoint={}, module={}, shape={})'.format(
            self.ownerL.to_string(), self.ownerR.to_string(), self.fixedpoint, self.xL.module, self.xL.shape)

    def assign(self, other: SharedPairBase) -> tf.Operation:

        if self.fixedpoint != other.fixedpoint:
            other = other.dup_with_precision(new_fixedpoint=self.fixedpoint)
        with tf.device(self.ownerL):
            assign_opL = self.xL.assign(other.xL)
        with tf.device(self.ownerR):
            assign_opR = self.xR.assign(other.xR)
        return tf.group(assign_opL, assign_opR)

    def load_from_tf_tensor(self, x: tf.Tensor, const=False):
        """

        :param x: tf.Tensor
        :return:
        """
        if const:
            # super(SharedVariablePair, self).from_numpy(x)
            super(SharedVariablePair, self).load_from_tf_tensor(x)
        else:
            with tf.device(self.ownerL):
                px = PrivateTensorBase(owner=self.ownerL, fixedpoint=self.fixedpoint)
                px.load_from_tf_tensor(x)
                u = px.to_SharedTensor()
                uL = u + u.ones_like()
                uR = -u.ones_like()
                self.xL = SharedVariable(initial_inner_value=uL.inner_value, module=uL.module)
                self.fixedpoint = px.fixedpoint
            with tf.device(self.ownerR):
                self.xR = SharedVariable(initial_inner_value=uR.inner_value, module=uR.module)

    def load_from_numpy(self, x: np.ndarray, const=False):
        if const:
            super(SharedVariablePair, self).load_from_numpy(x)
        else:
            with tf.device(self.ownerL):
                px = PrivateTensorBase(owner=self.ownerL)

                px.load_from_tf_tensor(tf.constant(x))
                xL = px.inner_value + tf.ones_like(px.inner_value)
                self.xL = SharedVariable(initial_inner_value=xL, module=px.module)
                self.fixedpoint = px.fixedpoint
            with tf.device(self.ownerR):
                xR = -tf.ones_like(xL)
                self.xR = SharedVariable(initial_inner_value=xR, module=px.module)
